#  Copyright (c) 2022 - Thumos - Jon Cavallie Mester
from typing import List


class SlidingWindow:
    _internal_list: List[bool]
    _window_length: int

    def __init__(self, window_length: int):
        self._internal_list = []
        self._window_length = window_length

    def add(self, result: bool):
        if len(self._internal_list) >= self._window_length:
            self._internal_list.pop(0)
        self._internal_list.append(result)

    def get_failure_count(self):
        return sum(map(lambda result: not result, self._internal_list))



#  Copyright (c) 2022 - Thumos - Jon Cavallie Mester

import unittest

from src.resiliens.fallback.Fallback import WithFallback


def fallback_for_function(foo, bar):
    return foo, bar


@WithFallback(fallback_for_function)
def failing_function(foo, bar):
    raise Exception


@WithFallback(fallback_for_function)
def succeeding_function(foo, bar):
    return bar, foo


class TestFallback(unittest.TestCase):
    fallback_got_called: bool

    def setUp(self) -> None:
        self.fallback_got_called = False

    def method_fallback(self):
        self.fallback_got_called = True

    @WithFallback(method_fallback)
    def failing_method_with_fallback(self):
        raise Exception()

    @WithFallback(method_fallback)
    def succeeding_method_with_fallback(self):
        return True

    def test_methodFails_fallbackGetsCalled(self):
        self.failing_method_with_fallback()
        self.assertTrue(self.fallback_got_called)

    def test_methodSucceeds_fallbackIsNotCalled(self):
        res = self.succeeding_method_with_fallback()
        self.assertTrue(res)
        self.assertFalse(self.fallback_got_called)

    def test_fallbackWithIOErrorAsExpected_regularExceptionIsThrown_fallbackNotCalled(
            self):

        @WithFallback(fallback=self.method_fallback,
                      expected_exception=IOError)
        def bogus_function(_self):
            raise Exception()

        self.assertRaises(Exception, bogus_function)

    @WithFallback(fallback=method_fallback, expected_exception=IOError)
    def bogus_function(self):
        raise IOError()

    def test_fallbackWithIOErrorAsExpected_IOErrorIsThrown_fallbackIsCalled(
            self):
        self.bogus_function()
        self.assertTrue(self.fallback_got_called)

    def test_functionFails_fallbackIsCalled(self):
        foo, bar = 1, 2
        foo2, bar2 = failing_function(foo, bar)
        self.assertEqual(bar, bar2)
        self.assertEqual(foo, foo2)

    def test_functionSucceeds_fallbackIsNotCalled(self):
        foo, bar = 1, 2
        foo2, bar2 = failing_function(foo, bar)
        self.assertEqual(bar, bar2)
        self.assertEqual(foo, foo2)



#  Copyright (c) 2022 - Thumos - Jon Cavallie Mester

from functools import wraps
from inspect import isgeneratorfunction
from typing import Callable, Type, Any


class FallbackClass:
    fallback: Callable
    fallback_exception: Callable
    expected_exception: Exception

    def __init__(self,
                 fallback: Callable = None,
                 fallback_exception: Callable = None,
                 expected_exception: Type[BaseException] = Exception):
        self.fallback = fallback
        self.fallback_exception = fallback_exception
        self._expected_exception = expected_exception

        if not self.fallback and not self.fallback_exception:
            raise TypeError(
                "Fallback decorator requires either a \"fallback\" or \"fallback_exception\" argument and "
                "neither was given.")
        if self.fallback and not callable(self.fallback):
            raise TypeError(
                "Argument \"fallback\" must be callable (i.e. a function)")
        if self.fallback_exception and not callable(self.fallback_exception):
            raise TypeError(
                "Argument \"fallback_exception\" must be callable (i.e. a function)"
            )

    def __call__(self, decorated_function):
        return self.decorate(decorated_function)

    @staticmethod
    def call(func, *args, **kwargs) -> Any:
        return func(*args, **kwargs)

    @staticmethod
    def call_generator(func, *args, **kwargs):
        for el in func(*args, **kwargs):
            yield el

    def decorate(self, function_to_decorate: Callable = None) -> Callable:
        call = self.call_generator if isgeneratorfunction(
            function_to_decorate) else self.call

        @wraps(function_to_decorate)
        def wrapper(*args, **kwargs):
            return self.try_catch_fallback(call, function_to_decorate, *args,
                                           **kwargs)

        return wrapper

    def try_catch_fallback(self, call, function_to_decorate, *args, **kwargs):
        try:
            return call(function_to_decorate, *args, **kwargs)
        except Exception as e:
            if issubclass(e.__class__, self._expected_exception):
                if self.fallback_exception:
                    return self.fallback_exception(e, *args, **kwargs)
                else:
                    return call(self.fallback, *args, **kwargs)
            else:
                raise e


def WithFallback(fallback: Callable = None,
                 fallback_exception: Callable = None,
                 expected_exception: Type[BaseException] = Exception):
    """
    Provide a fallback function for the decorated function in case an exception is thrown. NOTE: The fallback
    function must take the same number of arguments as the decorated function. Optionally, the fallback function may
    take the thrown exception as a first argument, followed by the arguments of the decorated function - in that
    case, use the fallback_exception parameter to specify the fallback function.
    :param fallback: Reference to the fallback function - note that it needs to same function signature as the decorated function.
    :param fallback_exception: Reference to the fallback function that takes the thrown exception as an additional argument
    - note that it needs to same function signature as the decorated function.
    :param expected_exception: Exception class you want to use fallback for. Default is the base Exception, but you may only want
    to use the fallback for, say, IOError and in that case you should specify it here.
    """
    return FallbackClass(fallback=fallback,
                         fallback_exception=fallback_exception,
                         expected_exception=expected_exception)



import time
import unittest

from src.resiliens.circuit_breaker import CircuitBreaker
from src.resiliens.circuit_breaker import CircuitBreakerStatus


class TestCircuitBreaker(unittest.TestCase):
    MAX_ATTEMPTS: int = 5
    WINDOW_SIZE: int = 10

    failed_count: int
    successful_count: int
    circuit_breaker: CircuitBreaker

    def setUp(self) -> None:
        self.failed_count = 0
        self.successful_count = 0

    @CircuitBreaker(failures=MAX_ATTEMPTS, sliding_window_size=WINDOW_SIZE)
    def fake_successful_http_call(self):
        self.successful_count += 1
        return True

    @CircuitBreaker(failures=MAX_ATTEMPTS, sliding_window_size=WINDOW_SIZE)
    def fake_failed_http_call(self):
        self.failed_count += 1
        raise Exception()

    def test_slidingWindow50Percent_circuitBreakerOpensAfter5Failures(self):
        should_fail = True
        reset_time = 1000

        @CircuitBreaker(failures=self.MAX_ATTEMPTS,
                        reset_timeout=reset_time,
                        sliding_window_size=self.WINDOW_SIZE)
        def test_func():
            if should_fail:
                self.failed_count += 1
                raise Exception()
            self.successful_count += 1

        for i in range(0, self.MAX_ATTEMPTS * 2):
            try:
                test_func()
            except Exception as ignored:
                pass
        expected = self.MAX_ATTEMPTS
        actual = self.failed_count
        self.assertEqual(expected, actual)

        should_fail = False
        time.sleep(reset_time / 1000)

        test_func()
        expected = 1
        actual = self.successful_count
        self.assertEqual(expected, actual)

    def test_circuitBreakerIsGivenName_getNameProperty_returnsRightName(self):
        self.circuit_breaker = CircuitBreaker(name="CB1")

        expected = "CB1"
        actual = self.circuit_breaker.name

        self.assertEqual(expected, actual)

    def test_circuitBreakerIsGivenMaxAttempts_getMaxAttemptsProperty_returnsRightMaxAttempts(
            self):
        self.circuit_breaker = CircuitBreaker(failures=100)

        expected = 100
        actual = self.circuit_breaker.failure_threshold

        self.assertEqual(expected, actual)

    def test_circuitBreakerIsInitialized_circuitBreakerStatusIsClosed(self):
        self.circuit_breaker = CircuitBreaker()

        expected = CircuitBreakerStatus.closed
        actual = self.circuit_breaker.status

        self.assertEqual(expected, actual)



#  Copyright (c) 2022 - Thumos - Jon Cavallie Mester


class _CircuitBreakerStatus:
    _CLOSED: str = 'CLOSED'
    _OPEN: str = 'OPEN'
    _HALF_OPEN: str = 'HALF_OPEN'

    @property
    def closed(self) -> str:
        return self._CLOSED

    @property
    def open(self) -> str:
        return self._OPEN

    @property
    def half_open(self) -> str:
        return self._HALF_OPEN

    def is_valid_status(self, status: str):
        return status in (self._OPEN, self._HALF_OPEN, self._CLOSED)


CircuitBreakerStatus = _CircuitBreakerStatus()
